
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

#include "src/float_cast.h"

#define	BUFFER_SIZE		(1<<16)
#define ARRAY_LEN(x)	((sizeof (x))/ (sizeof ((x)[0])))

enum
{	CLIP_POSITIVE_BAD	=	1,
	CLIP_NEGATIVE_BAD	=	2
} ;


static short
ar_round_1 (float a) ;
short ar_round_2 (float a) ;
static short ar_round_3 (float a) ;

short
ar_round (float a)
{
	a = a * 32768.0 + 65537.5;
	return (short)a;
} /* ar_round */





static void
f2lei_array (float *src, int *dest, int count, float normfact)
{	int			value ;
	
	while (count)
	{	count -- ;
		value = lrintf (src [count] * normfact) ;
		dest [count] = value ;
		} ;
} /* f2lei_array */

static void
new_f2lei_array (float *src, int *dest, int count, float normfact)
{	int				value ;
	
	while (count)
	{	count -- ;
		value = lrint (src [count] * normfact) ;

		if (value > (1.0 * 0x7FFFFFFF))
			value = 0x7FFFFFFF ;
		if (value < (-8.0 * 0x10000000))
			value = 0x80000000 ;

		dest [count] = value ;
		} ;
} /* new_f2lei_array */

static void
check_src_dest (float *src, int *dest, int count)
{	int		k, over_count, clip_flags ;

	over_count = 0 ;
	for (k = 0 ; k < count ; k++)
		if (fabs (src [k]) > 1.0)
			over_count ++ ;

	if (over_count < 0)
	{	puts ("Error : No clipping.") ;
		exit (1) ;
		} ;

	if ((1.0 * over_count) / count > 0.7)
	{	puts ("Error : Too much clipping.") ;
		exit (1) ;
		} ;
	
	/* Source is OK, now check dest. */
	clip_flags = 0 ;
	for (k = 0 ; k < count ; k++)
	{	if (src [k] > 0.0 && dest [k] < 0)
			clip_flags |= CLIP_POSITIVE_BAD ;
		if (src [k] < 0 && dest [k] > 0)
			clip_flags |= CLIP_NEGATIVE_BAD ;
		} ;

	if (clip_flags & CLIP_POSITIVE_BAD)
		puts ("positive bad") ;
	if (clip_flags & CLIP_NEGATIVE_BAD)
		puts ("negative bad") ;
		
} /* check_src_dest */

int
main (void)
{	static float	src [BUFFER_SIZE] ;
	static int		dest [BUFFER_SIZE] ;
	
	struct timeval start_time, end_time, time_diff ;
	double old_diff, new_diff ;
	int k ;
	
	puts ("\n") ;

	for (k = 0 ; k < ARRAY_LEN (src) ; k++)
		src [k] = 2.0 * sin (4 * 2 * M_PI * k / ARRAY_LEN (src)) ;
	
	f2lei_array (src, dest, ARRAY_LEN (src), 8.0 * 0x10000000) ;
	check_src_dest (src, dest, ARRAY_LEN (src)) ;

	/*------------------------------------------------------------*/

	gettimeofday (&start_time, NULL) ;
 	for (k = 0 ; k < 500 ; k ++)
 		f2lei_array (src, dest, ARRAY_LEN (src), 8.0 * 0x10000000) ;
	gettimeofday (&end_time, NULL) ;

	timersub (&end_time, &start_time, &time_diff) ;
	old_diff = time_diff.tv_sec * 1000 + time_diff.tv_usec / 1000.0 ;
	printf ("Old version : %.3f milliseconds.\n", old_diff) ;

	/*------------------------------------------------------------*/

	gettimeofday (&start_time, NULL) ;
	for (k = 0 ; k < 500 ; k ++)
		new_f2lei_array (src, dest, ARRAY_LEN (src), 8.0 * 0x10000000) ;
	gettimeofday (&end_time, NULL) ;

	timersub (&end_time, &start_time, &time_diff) ;
	new_diff = time_diff.tv_sec * 1000 + time_diff.tv_usec / 1000.0 ;

	printf ("New version : %.3f milliseconds.\n", new_diff) ;

	printf ("\n\nNew version runs at %.1f%% the speed of the old version.\n\n", 100.0 * old_diff / new_diff) ;
	
	puts ("pitchscaler in LADSPA") ;

	return 0 ;
} /* main */

/*
** Do not edit or modify anything in this comment block.
** The arch-tag line is a file identity tag for the GNU Arch 
** revision control system.
**
** arch-tag: 1f750ed1-539a-4d5f-a2b1-ea49a2b87fc6
*/

