
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

#include "src/float_cast.h"

#define ARRAY_LEN(x) ((sizeof (x))/ (sizeof ((x)[0])))

/*-#undef	CPU_IS_BIG_ENDIAN
#define	CPU_IS_BIG_ENDIAN 0
-*/

#if 1

static void
f2les_array (float *src, short *dest, int count, float normfact)
{	
	unsigned char	*ucptr ;
	short			value ;
	
	ucptr = ((unsigned char*) dest) + 2 * count ;
	while (count)
	{	count -- ;
		ucptr -= 2 ;
		value = lrintf (src [count] * normfact) ;
		ucptr [0] = value ;
		ucptr [1] = value >> 8 ;
		} ;

} /* f2les_array */

static void
new_f2les_array (float *src, short *dest, int count, float normfact)
{	unsigned char	*ucptr ;
	int				value ;
	
	ucptr = ((unsigned char*) dest) + 2 * count ;
	while (count)
	{	count -- ;
		ucptr -= 2 ;
		value = lrint (src [count] * normfact) ;

		if (value > 32767)
			value = 32767 ;
		if (value < -32768)
			value = -32768 ;

		ucptr [0] = value ;
		ucptr [1] = value >> 8 ;
		} ;
} /* new_f2les_array */


int
main (void)
{	static float src [1<<16] ;
	static short dest [1<<16] ;
	
	struct timeval start_time, end_time, time_diff ;
	double old_diff, new_diff ;
	int k ;
	
	puts ("\n") ;

	gettimeofday (&start_time, NULL) ;
 	for (k = 0 ; k < 500 ; k ++)
 		f2les_array (src, dest, ARRAY_LEN (src), 32767.0) ;
	gettimeofday (&end_time, NULL) ;

	timersub (&end_time, &start_time, &time_diff) ;
	old_diff = time_diff.tv_sec * 1000 + time_diff.tv_usec / 1000.0 ;
	printf ("Old version : %.3f milliseconds.\n", old_diff) ;
	
	/*------------------------------------------------------------*/

	gettimeofday (&start_time, NULL) ;
	for (k = 0 ; k < 500 ; k ++)
		new_f2les_array (src, dest, ARRAY_LEN (src), 32767.0) ;
	gettimeofday (&end_time, NULL) ;

	timersub (&end_time, &start_time, &time_diff) ;
	new_diff = time_diff.tv_sec * 1000 + time_diff.tv_usec / 1000.0 ;

	printf ("New version : %.3f milliseconds.\n", new_diff) ;

	printf ("\nNew version runs at %.1f%% the speed of the old version.\n\n", 100.0 * old_diff / new_diff) ;

	return 0 ;
} /* main */


#else

int
main (void)
{	short   short_val, output_val ;
	float 	float_val ;
	
	
	for (short_val = 0 ; short_val < 0x7FF0 ; short_val ++)
	{	
		float_val = short_val / (1.0 * 0x8000) ;
	
		output_val = lrintf (float_val * 0x7FFF) ;
		
		if (short_val != output_val)
		{	printf ("Original : %d\n", short_val) ;
			printf ("As float : %f\n", float_val) ;
			printf ("Output   : %d\n", output_val) ;
			exit (1) ;
			} ;
		} ;
	

	return 0 ;
} /* main */

#endif
/*
** Do not edit or modify anything in this comment block.
** The arch-tag line is a file identity tag for the GNU Arch 
** revision control system.
**
** arch-tag: 5437a6d7-a04a-402c-b135-aab3f9d01bea
*/

