#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#define ITS2RAW_LITTLE_ENDIAN 1234
#define ITS2RAW_BIG_ENDIAN 4321


/* --- uncomment one of these, the 1st for i386, the 2nd for ppc :) */
/* #define ITS2RAW_BYTE_ORDER ITS2RAW_LITTLE_ENDIAN */
/* #define ITS2RAW_BYTE_ORDER ITS2RAW_BIG_ENDIAN */


#if !defined(ITS2RAW_BYTE_ORDER)
#if defined(__NetBSD__)
#if _BYTE_ORDER == _LITTLE_ENDIAN
#define ITS2RAW_BYTE_ORDER ITS2RAW_LITTLE_ENDIAN
#else /* _BIG_ENDIAN */
#define ITS2RAW_BYTE_ORDER ITS2RAW_BIG_ENDIAN
#endif
#else /* !defined(__NetBSD__) */
#error Please define ITS2RAW_BYTE_ORDER based on your architecture.
#endif
#endif

#define buffer_size 1024

#define XCHG_SHORT(x) ((((x)&0xFF)<<8) | (((x)>>8)&0xFF))
#define XCHG_LONG(x) ((((x)&0xFF)<<24) | \
        (((x)&0xFF00)<<8) | \
	(((x)&0xFF0000)>>8) | \
	(((x)>>24)&0xFF))
#if ITS2RAW_BYTE_ORDER == ITS2RAW_LITTLE_ENDIAN
#define LE_SHORT(x) (x)
#define LE_LONG(x) (x)
#define BE_SHORT(x) XCHG_SHORT(x)
#define BE_LONG(x) XCHG_LONG(x)
#else /* _BIG_ENDIAN */
#define BE_SHORT(x) (x)
#define BE_LONG(x) (x)
#define LE_SHORT(x) XCHG_SHORT(x)
#define LE_LONG(x) XCHG_LONG(x)
#endif

#define RD(x, y) \
	if (x != read(fd, y, x)) { \
		printf("Error reading Impulse Tracker sample.\n"); \
		exit(1); \
	}

#define WR(x, y) \
	snprintf(tmp, buffer_size - 1, x, y); \
	tmp[buffer_size - 1] = '\000'; \
	write(fd, tmp, strlen(tmp))

#define READ_SHORT(thing) RD(2, &tmpshort); thing = LE_SHORT(tmpshort)
#define READ_LONG(thing) RD(4, &tmplong); thing = LE_LONG(tmplong)



char *basename(char *name)
{
	int l = strlen(name);
	char *retval = strdup(name);

	if (retval[l - 4] == '.') {
		retval[l - 4] = '\000';
	}

	return retval;
}

#define FLAGS_SAMPLE		(1<<0)	/* vs header only */
#define FLAGS_16BIT		(1<<1)	/* vs 8bit */
#define FLAGS_STEREO		(1<<2)	/* vs mono */
#define FLAGS_RESERVED		(1<<3)
#define FLAGS_LOOP		(1<<4)
#define FLAGS_SUSTAIN		(1<<5)
#define FLAGS_LOOP_PINGPONG	(1<<6)	/* vs loop forward */
#define FLAGS_SUSTAIN_PINGPONG	(1<<7)	/* vs sustain forward */

#define VIBRATO_SINE		0
#define VIBRATO_RAMP		1
#define VIBRATO_SQUARE		2
#define VIBRATO_RANDOM		3

#define CONVERT_SIGNED		(1<<0)	/* vs unsigned */
#define CONVERT_BE		(1<<1)	/* vs LE */
#define CONVERT_DELTA		(1<<2)	/* vs PCM */
#define CONVERT_BYTE_DELTA	(1<<3)
#define CONVERT_12BIT		(1<<4)

int main(int argc, char *argv[])
{
	int fd;
	char tmp[buffer_size];
	short int tmpshort;
	int tmplong;
	char file_name[13];
	char sample_name[26];
	unsigned char global_volume;
	unsigned char flags;
	unsigned char default_volume;
	unsigned char convert;
	unsigned char default_pan;
	int length;
	int loop_begin;
	int loop_end;
	int c5speed;
	int sustain_loop_begin;
	int sustain_loop_end;
	int sample_pointer;
	int length_in_bytes;
	void *data;
	char *vibrato_waveform_name[] = { "sine", "ramp", "square", "random" };
	char *bname;
	char *name = argv[1];
	unsigned char vibrato_speed;
	unsigned char vibrato_depth;
	unsigned char vibrato_waveform;
	unsigned char vibrato_rate;

	if (argc < 2) {
		printf("usage: %s [impulse tracker sample file]\n\n"
			"Reads an impulse tracker sample file and writes a\n"
			".raw audio data file and a .txt description.\n\n"
			"%s version 1.0 by Ben Collver\n"
			"http://netbsd.math.orst.edu/misc/its2raw.c\n\n",
			argv[0],
			argv[0]);
		exit(1);
	}

	fd = open(name, O_RDONLY);
	RD(4, tmp);
	if (0 != memcmp(tmp, "IMPS", 4)) {
		printf("This is not an Impulse Tracker sample file.\n");
		exit(1);
	}
	memset(file_name, 0, sizeof(file_name));
	RD(12, file_name);
	RD(1, tmp);
	RD(1, &global_volume);
	RD(1, &flags);
	RD(1, &default_volume);
	RD(26, sample_name);
	RD(1, &convert);
	RD(1, &default_pan);
	READ_LONG(length);
	READ_LONG(loop_begin);
	READ_LONG(loop_end);
	READ_LONG(c5speed);
	READ_LONG(sustain_loop_begin);
	READ_LONG(sustain_loop_end);
	READ_LONG(sample_pointer);
	RD(1, &vibrato_speed);
	RD(1, &vibrato_depth);
	RD(1, &vibrato_waveform);
	RD(1, &vibrato_rate);
	lseek(fd, sample_pointer, SEEK_SET);
	length_in_bytes = (length *
		((flags & FLAGS_16BIT) ? 2 : 1) *
		((flags & FLAGS_STEREO) ? 2 : 1));
	data = malloc(length_in_bytes);
	RD(length_in_bytes, data);
	close(fd);

	bname = basename(name);

	snprintf(tmp, sizeof(tmp) - 1, "%s.raw", bname);
	fd = open(tmp, O_CREAT|O_WRONLY, 0666);
	write(fd, data, length_in_bytes);
	close(fd);

	snprintf(tmp, sizeof(tmp) - 1, "%s.txt", bname);
	fd = open(tmp, O_CREAT|O_WRONLY, 0666);

	WR("source: %s\n", argv[1]);
	WR("file_name: %s\n", file_name);
	WR("sample_name: %s\n", sample_name);
	WR("global_volume: %d\n", global_volume);
	WR("default_volume: %d\n", default_volume);
	WR("default_pan: %d\n", default_pan);
	WR("flags_sample: %d\n", 0 != (flags & FLAGS_SAMPLE));
	WR("flags_16bit: %d\n", 0 != (flags & FLAGS_16BIT));
	WR("flags_stereo: %d\n", 0 != (flags & FLAGS_STEREO));
	WR("flags_reserved: %d\n", 0 != (flags & FLAGS_RESERVED));
	WR("flags_loop: %d\n", 0 != (flags & FLAGS_LOOP));
	WR("flags_sustain: %d\n", 0 != (flags & FLAGS_SUSTAIN));
	WR("flags_loop_pingpong: %d\n", 0 != (flags & FLAGS_LOOP_PINGPONG));
	WR("flags_sustain_pingpong: %d\n", 0 != (flags & FLAGS_SUSTAIN_PINGPONG));
	WR("convert_signed: %d\n", 0 != (convert & CONVERT_SIGNED));
	WR("convert_big_endian: %d\n", 0 != (convert & CONVERT_BE));
	WR("convert_delta: %d\n", 0 != (convert & CONVERT_DELTA));
	WR("convert_byte_delta: %d\n", 0 != (convert & CONVERT_BYTE_DELTA));
	WR("convert_12bit: %d\n", 0 != (convert & CONVERT_12BIT));
	WR("loop_begin: %d\n", loop_begin);
	WR("loop_end: %d\n", loop_end);
	WR("c5speed: %d\n", c5speed);
	WR("sustain_loop_begin: %d\n", sustain_loop_begin);
	WR("sustain_loop_end: %d\n", sustain_loop_end);
	WR("vibrato_speed: %d\n", vibrato_speed);
	WR("vibrato_depth: %d\n", vibrato_depth);
	WR("vibrato_rate: %d\n", vibrato_rate);
	WR("vibrato_waveform: %s\n", vibrato_waveform_name[vibrato_waveform]);
	if ((0 != (convert & CONVERT_DELTA)) ||
		(0 != (convert & CONVERT_BYTE_DELTA)) ||
		(0 != (convert & CONVERT_12BIT)))
	{
		snprintf(tmp, sizeof(tmp) - 1,
			"sox_warnings: %s%s%s\n",
			(0 != (convert & CONVERT_DELTA)) ? "delta " : "",
			(0 != (convert & CONVERT_BYTE_DELTA)) ? "byte_delta " : "",
			(0 != (convert & CONVERT_12BIT)) ? "12bit " : "");
		tmp[sizeof(tmp) - 1] = '\000';
		write(fd, tmp, strlen(tmp));
	}
	snprintf(tmp, sizeof(tmp) - 1,
		"convert_to_wav: sox %s %s %s -r %d %s%s.raw %s.wav\n",
		(0 == (flags & FLAGS_16BIT)) ? "-b" : "-w",
		(0 == (convert & CONVERT_SIGNED)) ? "-u" : "-s",
		(0 == (flags & FLAGS_STEREO)) ? "-c 1" : "-c 2",
		c5speed,
		(ITS2RAW_BYTE_ORDER == ITS2RAW_LITTLE_ENDIAN) &&
			!(convert & CONVERT_BE) ? "" : "-x ",
		bname,
		bname);
	tmp[sizeof(tmp) - 1] = '\000';
	write(fd, tmp, strlen(tmp));
	close(fd);
	exit(0);
}
/*
** Do not edit or modify anything in this comment block.
** The arch-tag line is a file identity tag for the GNU Arch 
** revision control system.
**
** arch-tag: 89c3b249-1ce5-4976-9b78-c6ced48ba20b
*/

